@@ -133,11 +133,12 @@
 **  _________________________________________________________________
 */
 
-static BOOL        g_bFirstTime          = TRUE;
 static conn_rec   *g_pCurrentConnection  = NULL;
 static const char *g_szCacheServerPort   = NULL;
 static char       *g_szCacheServerPath   = NULL;
 static server_rec *g_pServer             = NULL;
+static pool       *g_pPermanentPool      = NULL;
+static table      *g_pKeyTable           = NULL; 
 
 
 /*  _________________________________________________________________
@@ -489,39 +490,25 @@ static void init_Module(server_rec *s, pool *p)
 {
     char *szLogFile;
     char buf[MAX_STRING_LEN];
-    static int nStashStdin, nStashStderr, nSaveStdin, nSaveStderr;
     server_rec *pMainServer = s;
     piped_log *pl;
     char *cp;
     char *cp2;
-
-    /*  
-     *  skip nasty first time initialisation of Apache API
-     *  to avoid problems with SSL initialization
-     */
-    if (g_bFirstTime) {
-        g_bFirstTime = FALSE;
-
-        /*  remember the current STDIO fds  */
-        nStashStdin = dup(STDIN_FILENO);
-        /* Yes, I really do mean stdout here, because stderr is 
-           already a logfile at this time, so we have
-           to use stdout for stderr later */
-        nStashStderr = dup(STDOUT_FILENO); 
-
-        /*  and just return this first time  */
-        return;
-    }
+    int nSaveStderr;
 
     /*
-     *  reconnect STDIO fds for passphrase dialog
+     * Create our own global memory pool
+     * which survives server restarts
+     */
+    if (g_pPermanentPool == NULL)
+        g_pPermanentPool = ap_make_sub_pool(NULL);
+
+    /*
+     * Reconnect stderr to terminal (=stdout)
      */
-    nSaveStdin = dup(STDIN_FILENO);
-    close(STDIN_FILENO);
-    (void)dup(nStashStdin);
     nSaveStderr = dup(STDERR_FILENO);
     close(STDERR_FILENO);
-    (void)dup(nStashStderr);
+    (void)dup(STDOUT_FILENO);
 
     /*
      *  make sure the required SSL session cache is configured
@@ -636,16 +623,11 @@ static void init_Module(server_rec *s, pool *p)
      */
     LaunchCache(pMainServer, p);
 
-    /* 
-     *  Restore STDIO fds to their old values
+    /*
+     *  Reconnect stderr to error log
      */
-    close(STDIN_FILENO);
-    (void)dup(nSaveStdin);
-    close(nSaveStdin);
-
     close(STDERR_FILENO);
     (void)dup(nSaveStderr);
-    close(nSaveStderr);
 
     /*
      *  Announce mod_ssl and SSLeay in HTTP Server field
@@ -780,22 +762,7 @@ static void init_GetCertAndKey(server_rec *s, pool *p, SSLConfigRec *pConfig)
                         __SSLeay X509_get_default_cert_area(), 
                         pConfig->szKeyFile);
     /* Otherwise the path already contains the name of the certificate file */
-    if ((f = ap_pfopen(p, szPath, "r")) == NULL) {
-        ap_log_error(APLOG_MARK, APLOG_ERR, s,
-                     "mod_ssl: Can't open SSL server private key file %s: ", szPath);
-        ap_clear_pool(p);
-        exit(1);
-    }
-    pConfig->prsaKey = __SSLeay RSA_new();
-    if (!__SSLeay PEM_read_RSAPrivateKey(f, &pConfig->prsaKey, NULL)) {
-        ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, s,
-                     "mod_ssl: Error reading SSL server private key file %s "
-                     "(SSLeay error follows)", szPath);
-        ssl_log_ssleay2apache(s);
-        ap_clear_pool(p);
-        exit(1);
-    }
-    ap_pfclose(p, f);
+    init_GetPrivateKey(s, p, pConfig, szPath);
 
     /*
      * ???
@@ -812,6 +779,49 @@ static void init_GetCertAndKey(server_rec *s, pool *p, SSLConfigRec *pConfig)
                 pConfig->szCACertificateFile));
 }
 
+static void init_GetPrivateKey(server_rec *s, pool *p, SSLConfigRec *pConfig,
+                               const char *szKeyFile)
+{
+    RSA *prsaKey;
+    FILE *fp;
+
+    if (g_pKeyTable == NULL) {
+        if ((g_pKeyTable = ap_make_table(g_pPermanentPool, 10)) == NULL) {
+            ap_log_error(APLOG_MARK,APLOG_CRIT, s, 
+                         "mod_ssl: Can't allocate table for private keys");
+            ap_clear_pool(p);
+            exit(1);
+        }
+    }
+    prsaKey = (RSA *)ap_table_get(g_pKeyTable, szKeyFile);
+    if (prsaKey != NULL) {
+        pConfig->prsaKey = prsaKey;
+        return;
+    }
+    if ((fp = ap_pfopen(p, szKeyFile, "r")) == NULL) {
+        ap_log_error(APLOG_MARK,APLOG_CRIT, s,
+                     "mod_ssl: Can't open SSL server private key file %s",
+                     szKeyFile);
+        ap_clear_pool(p);
+        exit(1);
+    }
+
+    pConfig->prsaKey = RSA_new();
+    if (!__SSLeay PEM_read_RSAPrivateKey(fp, &pConfig->prsaKey, NULL)) {
+        ap_log_error(APLOG_MARK, APLOG_CRIT, s,
+                    "mod_ssl: Failed to read private key file %s", szKeyFile);
+        ssl_log_ssleay2apache(s);
+        ap_clear_pool(p);
+        exit(1);
+    }
+    ap_pfclose(p, fp);
+
+    ap_table_setn(g_pKeyTable, 
+                  ap_pstrdup(g_pPermanentPool, szKeyFile),
+                  (const char *)pConfig->prsaKey);
+    return;
+}
+
 static int init_VerifyFlags(SSLConfigRec *pConfig)
 {
     int nVerify = 0;
@@ -1146,19 +1156,6 @@ static int ssl_callback_SSLVerify(int ok, X509_STORE_CTX *ctx)
 
     SSLConfigRec *pConfig = mySrvConfig(g_pCurrentConnection->server);
 
-#if DEBUG_VERIFY
-    {
-        char *s;
-        s = __SSLeay X509_NAME_oneline(__SSLeay X509_get_subject_name(xs), NULL, 0);
-        if (s == NULL) {
-            __SSLeay ERR_print_errors_fp(pConfig->fileLogFile);
-            return(0);
-        }
-        fprintf(pConfig->fileLogFile, "depth=%d %s\n", depth, s);
-        free(s);
-    }
-#endif
-
     if (   error == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY 
         || error == X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE) {
         char *s;
@@ -1195,14 +1192,6 @@ static int ssl_callback_SSLVerify(int ok, X509_STORE_CTX *ctx)
 
 static int ssl_callback_NewSessionCache(SSL *pSSL, SSL_SESSION *pNew)
 {
-#if DEBUG_SESSIONS
-    int n;
-
-    fprintf(stderr, "ssl_callback_NewSessionCache(");
-    for (n = 0; n < pNew->session_id_length; ++n)
-        fprintf(stderr, "%02x", pNew->session_id[n]);
-    fprintf(stderr, ")\n");
-#endif
     SendSessionToServer(pNew);
     return 0;
 }
@@ -1211,14 +1200,6 @@ static SSL_SESSION *ssl_callback_GetSessionCache(
     SSL *pSSL, unsigned char *aucSessionID, int nLength, int *pCopy)
 {
     SSL_SESSION *pSession;
-#if DEBUG_SESSIONS
-    int n;
-
-    fprintf(stderr, "ssl_callback_GetSessionCache(");
-    for (n = 0; n < nLength; ++n)
-        fprintf(stderr, "%02x", aucSessionID[n]);
-    fprintf(stderr, ")\n");
-#endif
 
     *pCopy = 0;
     pSession = GetSessionFromServer(aucSessionID, nLength);
