@@ -101,7 +101,6 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
-#include <assert.h>
 #include <time.h>
 
 /* mod_ssl headers */
@@ -148,15 +147,19 @@ uchar *ReadThing(int nFD, int *pnLength)
     int nRead;
 
     nRead = saferead(nFD, &usLength, sizeof usLength);
-    assert(nRead == sizeof usLength);
-    if (!usLength) {
+    if (nRead != sizeof usLength || !usLength) {
+        *pnLength = 0;
+        return NULL;
+    }
+    if ((aucThing = malloc(usLength)) == NULL) {
         *pnLength = 0;
         return NULL;
     }
-    aucThing = malloc(usLength);
-    assert(aucThing);
     nRead = saferead(nFD, aucThing, usLength);
-    assert(nRead == usLength);
+    if (nRead != usLength) {
+        *pnLength = 0;
+        return NULL;
+    }
 
     *pnLength = usLength;
     return aucThing;
@@ -215,7 +218,8 @@ Cache *LocalCacheAdd(uchar * aucKey, int nKey, time_t tExpiresAt)
             nCacheMax = INIT_CACHE_SIZE;
             pCache = malloc(nCacheMax * sizeof *pCache);
         }
-        assert(pCache);
+        if (pCache == NULL)
+            return NULL;
     }
 
     pCache[n].aucKey = aucKey;
